==> unruled_api-fabric%2B1.21-1.4-sources.jar: mc/recraftors/unruled_api/rules/RuleBuilder.java <==
package mc.recraftors.unruled_api.rules;

import com.mojang.brigadier.arguments.ArgumentType;
import com.mojang.brigadier.arguments.IntegerArgumentType;
import mc.recraftors.unruled_api.impl.GameruleValidatorAdapter;
import mc.recraftors.unruled_api.utils.GameruleAccessor;
import mc.recraftors.unruled_api.utils.IGameruleAdapter;
import mc.recraftors.unruled_api.utils.IGameruleValidator;
import mc.recraftors.unruled_api.utils.ServerBoundAccessor;
import net.minecraft.class_1928;
import net.minecraft.class_1928.class_4311;
import net.minecraft.class_1928.class_4312;
import net.minecraft.class_1928.class_4313;
import net.minecraft.class_1928.class_4314;
import net.minecraft.class_1928.class_5198;
import net.minecraft.class_1928.class_5199;
import net.minecraft.class_7699;
import net.minecraft.server.MinecraftServer;
import org.jetbrains.annotations.NotNull;

import java.util.Optional;
import java.util.function.BiConsumer;
import java.util.function.Supplier;

import static java.util.Objects.requireNonNull;
import static net.minecraft.class_1928.*;

/**
 * Base class for building gamerules with all the Unruled API features directly accessible.<br>
 * All base implementations can be easily accessed via the {@link mc.recraftors.unruled_api.UnruledApi}
 * static methods, to avoid multiplicative class imports.
 * @param <T> The gamerule type
 * @param <V> The gamerule's value type
 */
public abstract class RuleBuilder <T extends class_1928.class_4315<T>, V> {
    protected final Supplier<ArgumentType<?>> argumentSupplier;
    protected final class_5199<T> screenVisitor;
    protected V initialValue;
    protected BiConsumer<MinecraftServer, T> changeCallback;
    protected IGameruleValidator<V> validator;
    protected IGameruleAdapter<V> adapter;
    protected class_7699 featureSet;
    protected boolean serverBound;

    protected RuleBuilder(Supplier<ArgumentType<?>> argumentSupplier, class_5199<T> screenVisitor, V initialValue) {
        this.argumentSupplier = requireNonNull(argumentSupplier);
        this.screenVisitor = requireNonNull(screenVisitor);
        this.initialValue = initialValue;
        this.changeCallback = (w, s) -> {};
        this.validator = IGameruleValidator::alwaysTrue;
        this.adapter = Optional::of;
        this.featureSet = class_7699.method_45397();
    }

    protected abstract @NotNull T ruleBuilder(class_4314<T> type);

    /**
     * Sets a change callback for the rule to be built.<br>
     * A change callback is triggered any time the gamerule's value gets changed.
     * @param changeCallback The change callback to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setChangeCallback(BiConsumer<MinecraftServer, T> changeCallback) {
        this.changeCallback = requireNonNull(changeCallback);
        return this;
    }

    /**
     * Sets a value validator for the rule to be built.<br>
     * A value validator is checked any time the gamerule's value is set.<br>
     * Best used along with a value adapter. See {@link #setAdapter}.
     * @param validator The value validator to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setValidator(IGameruleValidator<V> validator) {
        this.validator = requireNonNull(validator);
        return this;
    }

    /**
     * Sets a value adapter for the rule to be built.<br>
     * A value adapter is applied when a value gets invalidated by the rule's validator.<br>
     * As such, it requires a validator to be set to be relevant. See {@link #setValidator}.
     * @param adapter The value adapter to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setAdapter(IGameruleAdapter<V> adapter) {
        this.adapter = requireNonNull(adapter);
        return this;
    }

    /**
     * Sets a value validatorAdapter for the rule to be built.<br>
     * A value validatorAdapter works as both validator and adapter.<br>
     * Some basic implementation can be found in {@link mc.recraftors.unruled_api.impl}.
     * @param validatorAdapter The value validatorAdapter to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setValidatorAdapter(GameruleValidatorAdapter<V> validatorAdapter) {
        requireNonNull(validatorAdapter);
        this.validator = validatorAdapter;
        this.adapter = validatorAdapter;
        return this;
    }

    /**
     * Sets a featureSet for the rule to be built.<br>
     * A featureSet (or experimental feature set) can determine whether a rule is indeed loaded or not.<br>
     * This is a Minecraft core feature, and not inherently used by Unruled API.
     * @param featureSet The featureSet to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setFeatureSet(class_7699 featureSet) {
        this.featureSet = requireNonNull(featureSet);
        return this;
    }

    /**
     * Sets whether a built rule should be server-bound.<br>
     * A server-bound rule cannot be overriden, and only the server value will ever be provided.
     * @param serverBound Whether a built rule should be server-bound.
     * @return this
     */
    public final RuleBuilder<T, V> setServerBound(boolean serverBound) {
        this.serverBound = serverBound;
        return this;
    }

    /**
     * Builds a gamerule Type (or core builder), which can be then registered for the rule to exist in the game.
     * @return The built rule Type.
     */
    public class_4314<T> build() {
        class_4314<T> type = new class_4314<>(this.argumentSupplier, this::ruleBuilder, this.changeCallback, this.screenVisitor);
        ((ServerBoundAccessor)type).unruled_setServerBound(this.serverBound);
        return type;
    }

    /**
     * Registers the builder's build result with the provided name and category, and
     * returns the matching gamerule key.
     * @param name The name of the newly registered rule
     * @param category The category of the newly registered rule
     * @return The key of the newly registered rule
     */
    public class_4313<T> register(String name, class_5198 category) {
        return class_1928.method_8359(name, category, build());
    }

    public static class IntRuleBuilder extends RuleBuilder<class_4312, Integer> {
        public IntRuleBuilder(int initialValue) {
            super(IntegerArgumentType::integer, class_4311::method_27330, initialValue);
        }

        @Override
        @SuppressWarnings("unchecked")
        protected @NotNull class_4312 ruleBuilder(class_4314<class_4312> type) {
            class_4312 rule = new class_4312(type, super.initialValue);
            GameruleAccessor<Integer> accessor = ((GameruleAccessor<Integer>)rule);
            accessor.unruled_setValidator(super.validator);
            accessor.unruled_setAdapter(super.adapter);
            return rule;
        }
    }
}

==> unruled_api-fabric%2B1.21.3-1.4.1-sources.jar: mc/recraftors/unruled_api/rules/RuleBuilder.java <==
package mc.recraftors.unruled_api.rules;

import com.mojang.brigadier.arguments.ArgumentType;
import com.mojang.brigadier.arguments.IntegerArgumentType;
import mc.recraftors.unruled_api.impl.GameruleValidatorAdapter;
import mc.recraftors.unruled_api.utils.GameruleAccessor;
import mc.recraftors.unruled_api.utils.IGameruleAdapter;
import mc.recraftors.unruled_api.utils.IGameruleValidator;
import mc.recraftors.unruled_api.utils.ServerBoundAccessor;
import net.minecraft.class_1928;
import net.minecraft.class_1928.class_4311;
import net.minecraft.class_1928.class_4312;
import net.minecraft.class_1928.class_4313;
import net.minecraft.class_1928.class_4314;
import net.minecraft.class_1928.class_5198;
import net.minecraft.class_1928.class_5199;
import net.minecraft.class_7699;
import net.minecraft.server.MinecraftServer;
import org.jetbrains.annotations.NotNull;

import java.util.Optional;
import java.util.function.BiConsumer;
import java.util.function.Supplier;

import static java.util.Objects.requireNonNull;
import static net.minecraft.class_1928.*;

/**
 * Base class for building gamerules with all the Unruled API features directly accessible.<br>
 * All base implementations can be easily accessed via the {@link mc.recraftors.unruled_api.UnruledApi}
 * static methods, to avoid multiplicative class imports.
 * @param <T> The gamerule type
 * @param <V> The gamerule's value type
 */
public abstract class RuleBuilder <T extends class_1928.class_4315<T>, V> {
    protected final Supplier<ArgumentType<?>> argumentSupplier;
    protected final class_5199<T> screenVisitor;
    protected V initialValue;
    protected BiConsumer<MinecraftServer, T> changeCallback;
    protected IGameruleValidator<V> validator;
    protected IGameruleAdapter<V> adapter;
    protected class_7699 featureSet;
    protected boolean serverBound;

    protected RuleBuilder(Supplier<ArgumentType<?>> argumentSupplier, class_5199<T> screenVisitor, V initialValue) {
        this.argumentSupplier = requireNonNull(argumentSupplier);
        this.screenVisitor = requireNonNull(screenVisitor);
        this.initialValue = initialValue;
        this.changeCallback = (w, s) -> {};
        this.validator = IGameruleValidator::alwaysTrue;
        this.adapter = Optional::of;
        this.featureSet = class_7699.method_45397();
    }

    protected abstract @NotNull T ruleBuilder(class_4314<T> type);

    /**
     * Returns the rule builder's configured featureSet.
     * @return The rule builder's configured featureSet.
     */
    protected final class_7699 getFeatureSet() {
        return this.featureSet;
    }

    /**
     * Sets a change callback for the rule to be built.<br>
     * A change callback is triggered any time the gamerule's value gets changed.
     * @param changeCallback The change callback to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setChangeCallback(BiConsumer<MinecraftServer, T> changeCallback) {
        this.changeCallback = requireNonNull(changeCallback);
        return this;
    }

    /**
     * Sets a value validator for the rule to be built.<br>
     * A value validator is checked any time the gamerule's value is set.<br>
     * Best used along with a value adapter. See {@link #setAdapter}.
     * @param validator The value validator to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setValidator(IGameruleValidator<V> validator) {
        this.validator = requireNonNull(validator);
        return this;
    }

    /**
     * Sets a value adapter for the rule to be built.<br>
     * A value adapter is applied when a value gets invalidated by the rule's validator.<br>
     * As such, it requires a validator to be set to be relevant. See {@link #setValidator}.
     * @param adapter The value adapter to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setAdapter(IGameruleAdapter<V> adapter) {
        this.adapter = requireNonNull(adapter);
        return this;
    }

    /**
     * Sets a value validatorAdapter for the rule to be built.<br>
     * A value validatorAdapter works as both validator and adapter.<br>
     * Some basic implementation can be found in {@link mc.recraftors.unruled_api.impl}.
     * @param validatorAdapter The value validatorAdapter to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setValidatorAdapter(GameruleValidatorAdapter<V> validatorAdapter) {
        requireNonNull(validatorAdapter);
        this.validator = validatorAdapter;
        this.adapter = validatorAdapter;
        return this;
    }

    /**
     * Sets a featureSet for the rule to be built.<br>
     * A featureSet (or experimental feature set) can determine whether a rule is indeed loaded or not.<br>
     * This is a Minecraft core feature, and not inherently used by Unruled API.
     * @param featureSet The featureSet to be applied to a built rule.
     * @return this
     */
    public final RuleBuilder<T, V> setFeatureSet(class_7699 featureSet) {
        this.featureSet = requireNonNull(featureSet);
        return this;
    }

    /**
     * Sets whether a built rule should be server-bound.<br>
     * A server-bound rule cannot be overriden, and only the server value will ever be provided.
     * @param serverBound Whether a built rule should be server-bound.
     * @return this
     */
    public final RuleBuilder<T, V> setServerBound(boolean serverBound) {
        this.serverBound = serverBound;
        return this;
    }

    /**
     * Builds a gamerule Type (or core builder), which can be then registered for the rule to exist in the game.
     * @return The built rule Type.
     */
    public class_4314<T> build() {
        class_4314<T> type = new class_4314<>(this.argumentSupplier, this::ruleBuilder, this.changeCallback, this.screenVisitor, this.featureSet);
        ((ServerBoundAccessor)type).unruled_setServerBound(this.serverBound);
        return type;
    }

    /**
     * Registers the builder's build result with the provided name and category, and
     * returns the matching gamerule key.
     * @param name The name of the newly registered rule
     * @param category The category of the newly registered rule
     * @return The key of the newly registered rule
     */
    public class_4313<T> register(String name, class_5198 category) {
        return class_1928.method_8359(name, category, build());
    }

    public static class IntRuleBuilder extends RuleBuilder<class_4312, Integer> {
        public IntRuleBuilder(int initialValue) {
            super(IntegerArgumentType::integer, class_4311::method_27330, initialValue);
        }

        @Override
        @SuppressWarnings("unchecked")
        protected @NotNull class_4312 ruleBuilder(class_4314<class_4312> type) {
            class_4312 rule = new class_4312(type, super.initialValue);
            GameruleAccessor<Integer> accessor = ((GameruleAccessor<Integer>)rule);
            accessor.unruled_setValidator(super.validator);
            accessor.unruled_setAdapter(super.adapter);
            return rule;
        }
    }
}
